import fds.sdk.PAEngine
from fds.sdk.PAEngine.api import (
    pa_calculations_api,
    accounts_api,
    components_api,
    column_statistics_api,
    columns_api,
    currencies_api,
    dates_api,
    documents_api,
    frequencies_api,
    groups_api,
    pricing_sources_api,
    unlinked_pa_templates_api,
)
from fds.protobuf.stach.extensions.StachVersion import StachVersion
from fds.protobuf.stach.extensions.StachExtensionFactory import StachExtensionFactory
from fds.sdk.PAEngine.models import *
from dateutil.parser import parse
from toolz import first
from IPython.display import display
import pandas as pd
import time
import json
import urllib
from urllib.parse import urlparse, urlunparse
import logging 

# See configuration.py for a list of all supported configuration parameters.

# Examples for each supported authentication method are below,
# choose one that satisfies your use case.

# (Preferred) OAuth 2.0: FactSetOAuth2
# See https://github.com/FactSet/enterprise-sdk#oauth-20
# for information on how to create the app-config.json file
# See https://github.com/FactSet/enterprise-sdk-utils-python#authentication
# for more information on using the ConfidentialClient class

class PAEngineESDK:
    def __init__(
        self, fds_username_hyphen_serial: str, fds_apikey: str, display_as_sdk=True, turn_on_logging=False
    ):
        self.__api_instance = self.create_config(
            fds_username_hyphen_serial=fds_username_hyphen_serial, fds_apikey=fds_apikey,turn_on_logging=turn_on_logging
        )
        self.formatresults = FormatResults()
        self.lookups = LookUps(self.__api_client)
        self.parameters = Parameters(self.__api_client)
        self.display_as_sdk = display_as_sdk

    def create_config(self, fds_username_hyphen_serial, fds_apikey,turn_on_logging):
        self.__fds_config = fds.sdk.PAEngine.Configuration(
            username=fds_username_hyphen_serial, password=fds_apikey
        )

        #internal_url ="https://api.inhouse-cauth.factset.com"
        
        if turn_on_logging:
            logging.basicConfig(level=logging.DEBUG)
            self.__fds_config.debug=  True
         
        

        self.__api_client = fds.sdk.PAEngine.ApiClient(self.__fds_config)
        
        # self.__api_client.set_default_header('X-DataDirect-Auth-Token','585094dd1b82eb20')
        return pa_calculations_api.PACalculationsApi(self.__api_client)

    def cancel_calculation_by_id(self, calculation_id: str):
        '''
        https://github.com/factset/enterprise-sdk/blob/main/code/python/PAEngine/v3/docs/PACalculationsApi.md#cancel_calculation_by_id

        Cancel or Delete an Existing Calculation
        '''

        # Enter a context with an instance of the API client
        try:
            print(
                f"DELETE: {self.__fds_config.host}/analytics/engines/pa/v3/calculations/{calculation_id}"
            )
            results =  self.__api_instance.cancel_calculation_by_id_with_http_info(id=calculation_id)
            
            if results[1] == 204:
                print(f"Calculation Successfully Cancelled: {calculation_id}")

            return results[1]
        except fds.sdk.PAEngine.ApiException as e:
            print("Exception when calling PAEngine->cancel_calculation_by_id: %s\n" % e)

    def return_all_calculations(self):
        ''''
        Return the full dictionary of calculation requested within the last 12 hours 

        "data": {
            "d9e0eb21f0374c06956cfa6dbad4ba93": {
            "status": "Queued",
            "units": 1,
            "requestTime": "2024-06-25T19:47:06.149692Z",
            "lastPollTime": "2024-06-25T19:47:06.149692Z"
            },
        }

        '''

        response = self.get_all_calculations()
        results = response.to_dict()['data']

        while(response.to_dict()['meta']['pagination']['next_page']):
            response = self.get_all_calculations(int(response.to_dict()['meta']['pagination']['next_page']))
            results.update(response['data'].to_dict())

        
        return {'data': results} 


    def get_all_calculations(self, page_number: int = 1):
        '''
        https://github.com/factset/enterprise-sdk/blob/main/code/python/PAEngine/v3/docs/PACalculationsApi.md#get_all_calculations      
        
        Return All Calculations requested within the last 12 hours 

        {
            "data": {
                
                "d9e0eb21f0374c06956cfa6dbad4ba93": {
                    "lastPollTime": "2024-06-25T19: 47: 06.149692Z",
                    "requestTime": "2024-06-25T19: 47: 06.149692Z",
                    "status": "Queued",
                    "units": 1
                }
            },
            "meta": {
                "pagination": {
                    "next_page": "2",
                    "page_number": 1,
                    "page_size": 20,
                    "previous_page": None,
                    "total_calculations": 22,
                    "total_pages": 2
                }
            }
        }

        '''
        try:
            print(
                f"GET: {self.__fds_config.host}/analytics/engines/pa/v3/calculations?pageNumber={page_number}"
            )
            return self.__api_instance.get_all_calculations(page_number=page_number)
        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling PACalculationsApi->get_all_calculations: %s\n"
                % e
            )

    def get_calculation_parameters(self, calculation_id):
        '''
        https://github.com/factset/enterprise-sdk/blob/main/code/python/PAEngine/v3/docs/PACalculationsApi.md#get_calculation_parameters

        Return all arguments specified and default arguments for the specified calculation IDs.      
        '''
        
        try:
            # Get PA calculation parameters by id
            print(
                f"GET: {self.__fds_config.host}/analytics/engines/pa/v3/calculations/{calculation_id}"
            )
            return self.__api_instance.get_calculation_parameters (id=calculation_id)

        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling PACalculationsApi->get_calculation_parameters: %s\n"
                % e
            )

    def get_calculation_status_by_id(self, calculation_id: str):
        '''
        Check Calculation Status for the Specified calculation_id 
        
        '''
        try:
            print(
                f"GET: {self.__fds_config.host}/analytics/engines/pa/v3/calculations/{calculation_id}/status"
            )
            return self.__api_instance.get_calculation_status_by_id_with_http_info(
                calculation_id
            )
        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling Calculation Status->get_calculation_status_by_id_with_http_info: %s\n"
                % e
            )

    def get_calculation_unit_result_by_id(
        self, calculation_id: str, unit_id: str
    ):
        '''        
        https://github.com/factset/enterprise-sdk/blob/main/code/python/PAEngine/v3/docs/PACalculationsApi.md#get_calculation_unit_result_by_id
        '''
        # Enter a context with an instance of the API client
        print(
            f"GET: {self.__fds_config.host}/analytics/engines/pa/v3/calculations/{calculation_id}/units/{FormatResults.encode_string(unit_id)}/result"
        )
        try:
            return self.__api_instance.get_calculation_unit_result_by_id_with_http_info(
                id=calculation_id, unit_id=unit_id
            )

        except fds.sdk.PAEngine.ApiException as e:
            print("Exception when calling PAEngine->get_calculation_result: %s\n" % e)



    def post_and_calculate(
        self,
        pa_calculation_parameters_root: PACalculationParametersRoot,
        x_fact_set_api_long_running_deadline: int = 10,
        max_stale: int = None,
    ):
        '''
        Request a New PA Calculation allowing FactSet to create the Calculation ID
        https://github.com/factset/enterprise-sdk/blob/main/code/python/PAEngine/v3/docs/PACalculationsApi.md#post_and_calculate
        
        '''
        dict_args = {
            "pa_calculation_parameters_root": pa_calculation_parameters_root,
            "x_fact_set_api_long_running_deadline": x_fact_set_api_long_running_deadline,
            **(
                {"cache_control": f"max-stale={max_stale}"}
                if max_stale is not None
                else {}
            ),
        }
        try:
            print("Making PA Engine Request")
            print(f"POST: {self.__fds_config.host}/analytics/engines/pa/v3/calculations")
            self.parameters.print_args(dict_args, self.display_as_sdk)
            return self.__api_instance.post_and_calculate_with_http_info(**dict_args)

        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling request PA calculation -> post_and_calculate: %s\n"
                % e
            )

    def put_and_calculate(
        self,
        id: str, 
        pa_calculation_parameters_root: PACalculationParametersRoot,
        x_fact_set_api_long_running_deadline: int = 10,
        max_stale: int = None,
    ):
        '''
           Create or Update PA calculation and run it. 
           https://github.com/factset/enterprise-sdk/blob/main/code/python/PAEngine/v3/docs/PACalculationsApi.md#put_and_calculate     
        '''

        dict_args = {
            "pa_calculation_parameters_root": pa_calculation_parameters_root,
            "x_fact_set_api_long_running_deadline": x_fact_set_api_long_running_deadline,
            **(
                {"cache_control": f"max-stale={max_stale}"}
                if max_stale is not None
                else {}
            ),
        }
        try:
            print("Making PA Engine Request")
            print(f"PUT: {self.__fds_config.host}/analytics/engines/pa/v3/calculations")
            self.parameters.print_args(dict_args, self.display_as_sdk)
            return self.__api_instance.put_and_calculate_with_http_info(id, **dict_args)

        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling request PA calculation -> post_and_calculate: %s\n"
                % e
            )

    def return_PAEngine_calculation(
        self,
        pa_calculation_parameters_root: PACalculationParametersRoot,
        x_fact_set_api_long_running_deadline: int = 10,
        max_stale: int = None,
        calculation_id = None 
    ):
        '''
        Request and Return Results
        '''
        
        (response, status_code, headers) = self.put_and_calculate(id=calculation_id,
            pa_calculation_parameters_root=pa_calculation_parameters_root,
            x_fact_set_api_long_running_deadline=x_fact_set_api_long_running_deadline,
            max_stale=max_stale,
        ) if calculation_id else self.post_and_calculate(
            pa_calculation_parameters_root=pa_calculation_parameters_root,
            x_fact_set_api_long_running_deadline=x_fact_set_api_long_running_deadline,
            max_stale=max_stale,
        )

        print(f"Status Code: {status_code}")

        if status_code == 201:
            return headers, [
                (
                    list(pa_calculation_parameters_root.to_dict()["data"].keys())[0],
                    {"json": response.get_response_201()},
                )
            ]
        if status_code == 200:
            status_response = response.get_response_200()
            return headers, self.return_PAEngine_calculation_response(
                calculation_id=status_response.to_dict()["data"]["calculationid"],
                response=status_response,
            )
        if status_code == 202:
            response, calculation_id, status_code = self.return_PAEngine_long_running_response(
                response.get_response_202()
            )
            if status_code == 200:
                return headers, self.return_PAEngine_calculation_response(calculation_id,response)
            
    def return_PAEngine_long_running_response(self, response):
        ''''
        Iterate and Poll for Long Running Requests 
        
        
        '''        
        print("Long Running")
        print("Response Body:")
        print(json.dumps(response.to_dict(), indent=5))
        calculation_id = response.to_dict()["data"]["calculationid"]
        response, status_code, headers = self.get_calculation_status_by_id(calculation_id)
              
        while status_code == 202:
            print(f"Status Code: {status_code}")            
            max_age = (
                headers["cache-control"].replace("max-age=", "")
                if "cache-contol" in headers.keys()
                else 5
            )
            print(
                f"Sleeping for {max_age} seconds."
                if max_age != "1"
                else f"Sleeping for {max_age} second."
            )
            time.sleep(int(max_age))
            response, status_code, headers = self.get_calculation_status_by_id(calculation_id)

        print(f"Status Code: {status_code}")
        print("Final Response Body:")
        print(json.dumps(response.to_dict(), indent=5))

        return response, calculation_id, status_code



    def return_PAEngine_calculation_response(self, calculation_id, response):
        try:
            return [
                self.update_unit(calculation_id=calculation_id, unit=unit)
                for unit in response["data"]["units"].items()
            ]
        except:
            print(f"Exception when Returning Results from PAEngine: {calculation_id}" )

    def update_unit(self, calculation_id, unit):
        dict_results = unit[1].to_dict()        
        if unit[1]["status"] == "Success":
            response, status_code, headers = self.get_calculation_unit_result_by_id(
                calculation_id=calculation_id, unit_id=unit[0]
            )
            dict_results["json"] = response

        return(unit[0],dict_results)

    def create_root(self):
        pa_calculation_parameters_root = PACalculationParametersRoot(
            data={
                "2FB": PACalculationParameters(
                    componentid="2A68C724C54E77AE3914AFCBAA264D37462863EDAA9F563A8C15E1552EAB33F5",
                    accounts=[PAIdentifier(id="lion:NY")],
                    benchmarks=[PAIdentifier(id="lion:IEV")],
                ),
                "Chars": PACalculationParameters(
                    componentid="3020CF30D791FA7A5A14A0BF3633DF917F11962A04A7857C62281D082A52F84A",
                    accounts=[PAIdentifier(id="lion:NY")],
                    benchmarks=[PAIdentifier(id="lion:IEV")],
                ),
            },
            meta=CalculationMeta(
                contentorganization="SimplifiedRow",
                stach_content_organization="SimplifiedRow",
                contenttype="Json",
                format="JsonStach",
            ),
        )

        return pa_calculation_parameters_root
class LookUps:

    def __init__(self, api_client):
        self.__api_client = api_client
        self.__FormatResults = FormatResults()
        self.__host = api_client.configuration.host
    def get_accounts(self, path: str = "client:"):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/AccountsApi.md#get_accounts
        Return all Accounts (.acct) Account Composites (.actm), and Sub-Directories for the specified FactSet Directory.
        The FactSet directory should start with Personal:, Client:, or Super_Client:
        {
            "data": {
                "accounts": [
                    "$$DEMO_PARSE_PW.ACCT",
                    "$$PERFORMANCE.ACCT",
                    "$$PW_DEMO_PARSE.ACCT"
                ],
                "directories": [
                    "$$aat",
                    "$$aat_test"
                ]
            }
        }
        
        '''
        try:
            directory_path = FormatResults.format_factset_directory(path)
            print(
                f"GET: {self.__host}/analytics/lookups/v3/accounts/{FormatResults.encode_string(directory_path)}"
            )
            
            results =  accounts_api.AccountsApi(self.__api_client).get_accounts(
                directory_path
            )
            return results

        except fds.sdk.PAEngine.ApiException as e:
            print("Exception when calling AccountsApi->get_accounts: %s\n" % e)

    def get_pa_column_statistics(self):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/ColumnStatisticsApi.md#get_pa_column_statistics
        This endpoint lists all the column statistics that can be applied to a PA column.
        example, this endpoint has no required or optional parameters
        {
            "data": {
                "eb9d6d91416e4224bacadc261787e56f": {
                    "name": "Active Weights"
                },
            }
        }
        '''
        try:
            print(
                f"GET: {self.__host}/analytics/engines/pa/v3/columnstatistics"
            )
            return column_statistics_api.ColumnStatisticsApi(
                self.__api_client
            ).get_pa_column_statistics()
        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling ColumnStatisticsApi->get_pa_column_statistics: %s\n"
                % e
            )

    def get_default_unlinked_pa_template_types(self):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/UnlinkedPATemplatesApi.md#get_default_unlinked_pa_template_types

        Return all Available Template Types IDs, names, and categories
        {
            "data": {
                "996E90B981AEE83F14029ED3D309FB3F03EC6E2ACC7FD42C22CBD5D279502CFD": {
                    "category": "Weights",
                    "name": "Weights"
                }
            }
        }

        '''
        try:
            print(
                f"GET: {self.__host}/analytics/engines/pa/v3/unlinked-templates/template-types"
            )
            return unlinked_pa_templates_api.UnlinkedPATemplatesApi(
                self.__api_client
            ).get_default_unlinked_pa_template_types()
        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling UnlinkedPATemplatesApi->get_default_unlinked_pa_template_types: %s\n"
                % e
            )

    def get_details_type(self, template_type_id: str):
        '''
       https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/UnlinkedPATemplatesApi.md#get_details_type  

       Returns the Name, Category, Columns, and Groups information for the specified template type id  

        '''
        try:
            print(
                f"GET: {self.__host}/analytics/engines/pa/v3/unlinked-templates/{template_type_id}"
            )
            return unlinked_pa_templates_api.UnlinkedPATemplatesApi(
                self.__api_client
            ).get_details_type(id=template_type_id)

        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling UnlinkedPATemplatesApi->get_default_unlinked_pa_template_types: %s\n"
                % e
            )
    
    def get_pa_columns(
        self, name: str = None, category: str = None, directory: str = None
    ):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/ColumnsApi.md#get_pa_columns

        Return all columns that can be added when requesting
        a new calculation or creating report or templated components.
        The universe of columns can be further refined by FactSet Directory (directory), Category (category), or Column Name (name).
        The directory argument should start with start with either Factset, Personal, Client, or  Super_Client.
        {
            "data":{
                "0E5D0648415D9B782078E4B048462633896171C21E919805EA94418F573E0162":{
                    "name":"Security Name",
                    "directory":"Factset",
                    "category":""
                },
                "05F05D9898A562CCE2F47D633C57510811552F0A911DC560ED6E3D22DECEFB95":{
                    "name":"Symbol",
                    "directory":"Factset",
                    "category":""
                }
            }
        }
        '''
        try:
            directory = (
                None
                if directory is None
                else directory[:-1] if directory[-1] == ":" else directory
            )

            args = {k: v for k, v in locals().items() if v is not None and k != "self"}
            url_part = "&".join(f"{k}={v}" for k, v in args.items())
            print(
                f"GET: {self.__host}/analytics/engines/pa/v3/columns?{url_part}"
                if url_part
                else f"GET: {self.__host}/analytics/engines/pa/v3/columns"
            )
            return columns_api.ColumnsApi(self.__api_client).get_pa_columns(**args)

        except fds.sdk.PAEngine.ApiException as e:
            print("Exception when calling ColumnsApi->get_pa_column_by_id: %s\n" % e)
    def get_pa_column_by_id(self, id: str):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/ColumnsApi.md#get_pa_column_by_id

        This endpoint returns the default settings of a PA column.
        {
            "data": {
                "defaultstatisticsids": [],
                "name": "Security Name",
                "directory": "Factset",
                "category": ""
            }
        }
        '''
        try:
            print(
                f"GET: {self.__host}/analytics/engines/pa/v3/columns/{id}"
            )
            return columns_api.ColumnsApi(self.__api_client).get_pa_column_by_id(
                id
            )
        except fds.sdk.PAEngine.ApiException as e:
            print("Exception when calling ColumnsApi->get_pa_column_by_id: %s\n" % e)



    def get_pa_components(self, document: str):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/ComponentsApi.md#get_pa_components

        Returns Name, Category, and ID for all components inside the specified PA3 Document
        The path will start with “factset:”, “client:”, “personal:”, or “super_client:”
        For a list of PA3 documents use the get_pa3_documents() function.
        Component IDs are required for requesting a new calculation or creating/modifying report or templated components.
        {
        "data":{
            "216D0AC369E1086647CF16F33435DCCF6C9984E799A941C8CF3F2D555AC899F9":{
                "name":"Weights",
                "category":"Weights / Exposures"
            },
            "BA3ED6AEBB4B3885954F9A762B93691CAB7D8721663D0AA419B462C4452BDD87":{
                "name":"Weights Difference",
                "category":"Weights / Exposures"
            }
        }
        }
        '''
        try:
            print(
                f"GET: {self.__host}/analytics/engines/pa/v3/components?document={document}"
            )
            return components_api.ComponentsApi(self.__api_client).get_pa_components(
                document=document
            )

        except fds.sdk.PAEngine.ApiException as e:
            print("Exception when calling ComponentsApi->get_pa_components: %s\n" % e)

    def get_pa_components_by_id(self, id: str):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/ComponentsApi.md#get_pa_component_by_id

        Returns the default settings for the specified Component ID
        '''
        try:
            print(
                f"GET: {self.__host}/analytics/engines/pa/v3/components/{id}"
            )
            return components_api.ComponentsApi(
                self.__api_client
            ).get_pa_component_by_id(id=id)

        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling ComponentsApi->get_pa_component_by_id: %s\n" % e
            )

    def get_currencies(self):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/CurrenciesApi.md#get_currencies

        Return all avaialable Currency ISO Names, and their codes.
        Set the currency to be applied to the calculation or temaplte
        {
            "data":{
                "AFN":{
                    "name":"Afghanistan Afghani"
                },
                "ALL":{
                    "name":"Albanian Lek"
                },
                "DZD":{
                    "name":"Algerian Dinar"
                }
            }
        }
        '''
        try:
            # Get currencies
            # example, this endpoint has no required or optional parameters
            print(f"GET: {self.__host}/analytics/lookups/v3/currencies")
            return currencies_api.CurrenciesApi(self.__api_client).get_currencies()

        except fds.sdk.PAEngine.ApiException as e:
            print("Exception when calling CurrenciesApi->get_currencies: %s\n" % e)

    def convert_pa_dates_to_absolute_format(
        self, end_date: str, component_id: str, account: str, start_date: str = ""
    ):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/DatesApi.md#convert_pa_dates_to_absolute_format

        This function will validate the relative date ranges passed to the PA Engine.
        Use the combination of the Portfolio and Component ID to return the Absolute Date Equivalent for the relative date (range).
        The FactSet Portfolio path should start with Personal:, Client:, or Super_Client:.
        end_date, component_id, and account are required arguments.
        Optionally, pass in a second (start date) date to ensure no issues with overlapping dates
        Cannot be used with a Fund or Benchmark

        {'enddate': '20210331', 'startdate': '20201223'}

        '''
        try:
            # Convert PA dates to absolute format
            # example passing only required values which don't have defaults set
            # and optional values

            url_part = "&".join(
                f"{k.replace('_','') }={v}"
                for k, v in locals().items()
                if v is not None and k != "self"
            )
            print(
                f"GET: {self.__host}/analytics/engines/pa/v3/dates?{url_part}"
            )
            return dates_api.DatesApi(
                self.__api_client
            ).convert_pa_dates_to_absolute_format(
                enddate=end_date,
                componentid=component_id,
                account=account,
                startdate=start_date,
            )

        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling DatesApi->convert_pa_dates_to_absolute_format: %s\n"
                % e
            )

    def get_pa3_documents(self, fds_directory: str = "client:"):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/DocumentsApi.md#get_pa3_documents

        Returns all sub-directories and PA3 Documents for the specified factset directory (path).
        The path will start with “factset:”, “client:”, “personal:”, or “super_client:”
        Results can be used as an input to the get_components() function.
        {
            "data":{
                "documents":[
                    "Academic Default Reports",
                    "Decision Analysis",
                    "Default"
                ]
            }
        }
        '''
        try:
            print(
                f"GET: {self.__host}/analytics/engines/pa/v3/documents/{FormatResults.encode_string(fds_directory)}"
            )
            # Get PA3 documents and sub-directories in a directory
            # example passing only required values which don't have defaults set
            return documents_api.DocumentsApi(self.__api_client).get_pa3_documents(
                fds_directory
            )

        except fds.sdk.PAEngine.ApiException as e:
            print("Exception when calling DocumentsApi->get_pa3_documents: %s\n" % e)

    def get_pa_frequencies(self):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/FrequenciesApi.md#get_pa_frequencies

        Returns a full list of available frequency identifiers that can be applied to a calculation, component, or report template.
        {
            "data":{
                "Single":{
                    "name":"Single"
                },
                "Daily":{
                    "name":"Daily"
                }
            }
        }
        '''
        try:
            print(f"GET: {self.__host}/analytics/engines/pa/v3/frequencies")
            return frequencies_api.FrequenciesApi(
                self.__api_client
            ).get_pa_frequencies()

        except fds.sdk.PAEngine.ApiException as e:
            print("Exception when calling FrequenciesApi->get_pa_frequencies: %s\n" % e)

    def get_pa_groups(self):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/GroupsApi.md#get_pa_groups

        Returns the name, directory, category, and ID of all available groups.

        {
            "data":{
                "10D46537D08F37A521A57FC77105FFE7CA8F336AAFDF876B09EAA384A6024ED0":{
                    "name":"3M Momentum",
                    "directory":"Client",
                    "category":"Client"
                },
                "07D67EFA4356CEA4CFE65F545BB6B158BB6A8356D7C5F89A7544CDAC4A0936F2":{
                    "name":"AGI Asset Class",
                    "directory":"Client",
                    "category":"Client"
                }
            }
        }
        '''
        try:
            # Get PA groups
            # example, this endpoint has no required or optional parameters
            print(f"GET: {self.__host}/analytics/engines/pa/v3/groups")
            return groups_api.GroupsApi(self.__api_client).get_pa_groups()

        except fds.sdk.PAEngine.ApiException as e:
            print("Exception when calling GroupsApi->get_pa_groups: %s\n" % e)

    def get_pa_grouping_frequencies(self):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/GroupsApi.md#get_pa_grouping_frequencies

        Retuns all Available grouping frequencies
        {
        "data": {
            "BeginningOfPeriod": {
                "name": "Beginning Of Period"
            },
            "EndOfPeriod": {
                "name": "End Of Period"
            }
        }
        '''
        try:
            # Get PA groups
            # example, this endpoint has no required or optional parameters
            print(
                f"GET: {self.__host}/analytics/engines/pa/v3/grouping-frequencies"
            )
            return groups_api.GroupsApi(self.__api_client).get_pa_grouping_frequencies()

        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling GroupsApi->get_pa_grouping_frequencies: %s\n"
                % e
            )

    def get_pa_pricing_sources(
        self, name: str = None, category: str = None, directory: str = None
    ):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/PricingSourcesApi.md#get_pa_pricing_sources

        Returns the Name, Directory, Category, and ID of all available pricing sources.
        {
        "data": {
            "06047C44E2CF586DE83CD47E8DC1461F5264225C4B655C51DC8333CCC6D24857": {
                "name": "LDI MSL",
                "directory": "Additional Security Masters",
                "category": ""
            }
        }
        '''
        try:
            args = {k: v for k, v in locals().items() if v is not None and k != "self"}
            url_part = "&".join(f"{k}={v}" for k, v in args.items())
            print(
                f"GET: https://api.factset.com/analytics/engines/pa/v3/pricing-sources?{FormatResults.encode_string(url_part)}"
                if url_part
                else f"GET: {self.__host}/analytics/engines/pa/v3/pricing-sources"
            )
            return pricing_sources_api.PricingSourcesApi(
                self.__api_client
            ).get_pa_pricing_sources(**args)

        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling Pricing Sources->get_pricing_sources: %s\n" % e
            )

    def get_unlinked_pa_templates(self, category: str = None, directory: str = None):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/UnlinkedPATemplatesApi.md#get_unlinked_pa_templates

        Returns a full list of unlinked PA templates stored in the specified FactSet directory.
        The directory path will start with “client:”, “personal:”, or “super_client:”.
        This function optionally allows to filter the unlinked pa template by either the directory or category.
        *Currenlty unlinked pa templates can only be saved in the personal directory
        *you can only pass in either a directory or category, but not both
        {
            "630156ba0bef442c9f80a4eddf990a03":{
                "category":"Weights",
                "description":"Overlap Report",
                "name":"Personal:630156ba0bef442c9f80a4eddf990a03.UNLINKED_PA_TEMPLATE"
            }
        }
        '''
        try:
            
            if category:
                url_args = f"?category={category}"
                req_args = {"category": category}
            else:
                arg_dir = directory if directory else "personal:"
                url_args = f"?directory={FormatResults.encode_string(arg_dir)}"
                req_args = {"directory": arg_dir}

            print(
                f"GET: {self.__host}/analytics/engines/pa/v3/unlinked-templates{url_args}"
            )
            return unlinked_pa_templates_api.UnlinkedPATemplatesApi(
                self.__api_client
            ).get_unlinked_pa_templates(**req_args)
        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling UnlinkedPATemplatesApi->get_unlinked_pa_templates: %s\n"
                % e
            )

    def get_unlinked_pa_templates_by_id(self, template_id: str):
        '''
        https://github.com/FactSet/enterprise-sdk/tree/main/code/python/PAEngine/v3/docs/UnlinkedPATemplatesApi.md#get_unlinked_pa_templates_by_id

        Return Detail information for the specifed unlinked template id
        
        '''
        try:
            print(
                f"GET: {self.__host}/analytics/engines/pa/v3/unlinked-templates/{template_id}"
            )

            return unlinked_pa_templates_api.UnlinkedPATemplatesApi(
                self.__api_client
            ).get_unlinked_pa_templates_by_id(template_id)
        except fds.sdk.PAEngine.ApiException as e:
            print(
                "Exception when calling UnlinkedPATemplatesApi->get_unlinked_pa_templates_by_id: %s\n"
                % e
            )
class FormatResults:
    @staticmethod
    def display_results(
        results, display_as_dataframe, display_as_json, max_rows_returned
    ):
        
        if display_as_dataframe:
            FormatResults.truncate_and_display_tables(
                FormatResults.return_lookup_results_as_dataframe(results["data"]),
                max_rows_returned,
            )

        if display_as_json:
            print(json.dumps(results.to_dict(), indent=5))

    @staticmethod
    def encode_string(url: str):
        return urllib.parse.quote(url, safe="")

    staticmethod

    def format_dataframe(df):
        '''Reindex and Center Headers'''
        header_align = "center"
        text_align = "left"
        return (df.style.hide(axis="index")).set_table_styles(
            [
                dict(selector="td", props=[("text-align", text_align)]),
                dict(selector="th", props=[("text-align", header_align)]),
            ]
        )

    staticmethod

    def format_factset_directory(factset_dir):
        fds_dir = factset_dir[:-1] if factset_dir[-1] == "\\" else factset_dir
        return (
            f"{fds_dir}:"
            if ":" not in fds_dir
            else f"{fds_dir}/" if fds_dir[-1] != "/" else fds_dir
        )

    staticmethod
    def return_calculation_results_as_dataframe(result):
        """Convert Row Based Stach into DataFrame (List)"""
        stachBuilder = StachExtensionFactory.get_row_organized_builder(
            StachVersion.V2
        )
        stachExtension = stachBuilder.set_package(result).build()
        dataFramesList = stachExtension.convert_to_dataframe()
        return dataFramesList
        
    staticmethod
    def return_directory_category_dictionary(
        pa_dictionary, is_tuple=False, key_value="", display_value="name"
    ):
        results = {}
        results[""] = {"": [("", "")]}
        for k, v in pa_dictionary.items():
            category = v["category"] if not v["category"] == "" else "uncategorized"
            directory = v["directory"] if not v["directory"] == "" else "No Directory"
            if key_value:
                component_display = (
                    f"{v[display_value]}",
                    (
                        f"{v[display_value]} ({k})",
                        (
                            (directory, category, v[display_value], k, key_value)
                            if not is_tuple
                            else (directory, category, v[display_value], k, [])
                        ),
                    ),
                )
            else:
                component_display = (
                    f"{v[display_value]}",
                    (
                        f"{v[display_value]} ({k})",
                        (
                            (directory, category, v[display_value], k)
                            if not is_tuple
                            else (directory, category, v[display_value], k, [])
                        ),
                    ),
                )
            if directory in results:
                if category in results[directory]:
                    results[directory][category].append(component_display)
                else:
                    results[directory][category] = [("", "")] + [component_display]
            else:
                results[directory] = {"": [("", "")]}
                results[directory][category] = [("", "")] + [component_display]

        return dict(sorted(results.items()))

    staticmethod

    def return_categorized_dictionary(
        pa_dictionary, is_tuple=False, key_value="", display_value="name"
    ):
        '''Return Dictionary of Tuples Used to Populate the Category Item Box Widget Combination Below'''
        results = {}
        results[""] = [("", "")]
        for k, v in pa_dictionary.items():
            category = v.get('category','uncategorized')  #v["category"] if not v["category"] == "" else "uncategorized"
            d_value=v.get(display_value,f"No {display_value}")
            if key_value:
                                
                component_display = (
                    f"{d_value}",
                    (
                        f"{d_value} ({k})",
                        (
                            (category, d_value, k, key_value)
                            if not is_tuple
                            else (category, d_value, k, [])
                        ),
                    ),
                )
            else:            
                component_display = (
                    f"{d_value}",
                    (
                        f"{d_value} ({k})",
                        (
                            (category, d_value, k)
                            if not is_tuple
                            else (category, d_value, k, [])
                        ),
                    ),
                )
            if category in results:
                results[category].append(component_display)
            else:
                results[category] = [("", "")] + [component_display]

        return dict(sorted(results.items()))

    staticmethod

    def return_lookup_results_as_dataframe(results):
        '''format Analytics API LookUp Endpoints as Dataframes'''
        df = pd.DataFrame(
            [dict(v.to_dict(), **{"id": k}) for k, v in results.items()]
        ).rename(str.upper, axis="columns")
        return df

    staticmethod
    def return_all_calcs_results_as_dataframe(results):
        '''format Analytics API LookUp Endpoints as Dataframes'''
        df = pd.DataFrame(
            [dict(v, **{"id": k}) for k, v in results.items()]
        ).rename(str.upper, axis="columns")
        return df
    staticmethod

    def return_uncategorized_dictionary(pa_dictionary):
        '''Return TUPLE for uncategorized reference data from FactSet
        Append Empty Value in First Position'''
        results = []
        results.append(("", ""))

        for k, v in pa_dictionary.items():
            results.append((f"{v['name']} ({k})", k))

        return results

    @staticmethod
    def truncate_and_display_tables(df, max_rows_returned=20):
        '''Format Resulting Tables and Truncate Results to make viewing easier
        Returns if Truncated Dataframe and the Rows Skipped'''
        if df.shape[0] <= max_rows_returned or max_rows_returned < 0:
            display(FormatResults.format_dataframe(df))
        else:
            display(FormatResults.format_dataframe(df.head(max_rows_returned)))
            print(f"{str(len(df.index) - max_rows_returned)} Rows skipped!")


class Parameters:
    def __init__(self, api_client):
        self.__api_client = api_client

    def make_PA_calculation_column(self, column_id: str, statistics: list[str] = None):
        '''
        column_id = Column GUID
        statistics = List of 1 or More Columns Statistic GUIDs used when returning Aggregate values (exp Total Portfolio / Sector )

        '''
        return PACalculationColumn(id=column_id, statistics=statistics)

    def make_PA_calculation_data_sources(
        self,
        portfolio_pricing_sources: list[PACalculationPricingSource] = None,
        benchmark_pricing_sources: list[PACalculationPricingSource] = None,
        use_port_pricing_source_for_benchmark: bool = None,
    ):

        return PACalculationDataSources(
            portfoliopricingsources=portfolio_pricing_sources,
            benchmarkpricingsources=benchmark_pricing_sources,
            useportfoliopricingsourcesforbenchmark=use_port_pricing_source_for_benchmark,
        )

    def make_PA_calculation_group(self, group_id: str, group_frequency = None):
        '''
        group_id = PA Group GUID

        '''
        return PACalculationGroup(id=group_id,frequency=group_frequency)

    def make_PA_calculations_parameters(self, component_id: str, **kwargs):
        '''Returns PACalculationParameter Object
        component_id = PA Tile GUID
        '''
        return PACalculationParameters(componentid=component_id, **kwargs)

    def make_PA_calculation_parameters_root(self, data: str, meta=None):
        return (
            PACalculationParametersRoot(data=data, meta=meta)
            if meta
            else PACalculationParametersRoot(data=data)
        )

    def make_PA_date_parameters(
        self, end_date: str, frequency: str, start_date: str = ""
    ):
        if start_date:
            return PADateParameters(
                startdate=self.__check_date_format(start_date),
                enddate=self.__check_date_format(end_date),
                frequency=frequency,
            )
        else:
            return PADateParameters(
                enddate=self.__check_date_format(end_date),
                frequency=frequency,
            )

    def make_PA_identifier(self, PA_Identifier: str, holdings_mode: str = None):
        '''
        Used for creating both Portfolio and Benchmark objects
        PA_Identifier = specifies the full path to the account, composite, or benchmark. Each specified account(.acct) or composite(.actm) must have the appropriate postfix and each benchmark must have a BENCH: prefix.
        Optionally, each object can contain a __holdingsmode__ attribute set to __B&H__ (Buy and Hold), __TBR__ (Transaction based), or __OMS__ (Order Management System). The default holdings mode value is __B&H__.

        '''
        return PAIdentifier(id=PA_Identifier, holdingsmode=holdings_mode)

    def make_pa_pricing_source(self, pricing_source_id: str):
        return PACalculationPricingSource(id=pricing_source_id)
     
    def make_template_content_types(self, mandatory: list[str], optional: list[str], locked: list[str]):
        '''
        Define what fields can be modifed when requesting the calculation 
        Each array can contain any combination of the following values: "account," "benchmarks," "groups," "columns," "dates," "currencyisocode," and "componentdetail__".
        '''
        
        return TemplateContentTypes(mandatory=mandatory, optional=optional,locked=locked)


    def make_unlinked_template_parameters(
        self, directory: str, template_type_id: str, **kwargs
    ):
        '''
        Return UnlinkedPATemplateParameters object
        directory: location to save unlinked template.  The path will start with “client:”, “personal:”, or “super_client:”
        template_type_id: they component ID used to define the report type: get:/analytics/engines/pa/v3/unlinked-templates/template-types
        Returns an object used to define metadata for the new unlinked pa template.  This object will be passed into the unlinked_template_parameters_root object.
        '''
        return UnlinkedPATemplateParameters(
            directory=directory, template_type_id=template_type_id, **kwargs
        )

    def make_unlinked_template_parameters_root(self, data: str, meta=None):
        '''Returns Unlinked Template Parameters Root Object
        component_id = PA Tile GUID
        '''
        return UnlinkedPATemplateParametersRoot(data=data, meta=meta)

    def make_unlinked_template_update_parameters_root(self, data: str, meta=None):
        return UnlinkedPATemplateUpdateParametersRoot(data=data,meta=meta)

    def make_unlinked_template_update_parameters(self, **kwargs):
        return UnlinkedPATemplateUpdateParameters(**kwargs)

    def set_attribute(self, parameters, name: str, object):
        '''
        Add attribute to existing parameters
        '''
        parameters.set_attribute(name, object)

    # Additional Functions Used for Creating Parameters

    def __check_date_format(self, date_value: str):
        '''I have to account for "M" in the input.  The parser library treats -1AM as a valid date
        This function trys to determine if a date is in relative or absolute date format and then format the absolute dates to YYYYMMDD
        '''
        try:
            if not "M" in str(date_value).upper():
                return Parameters.format_absolute_date(date_value)
            else:
                return date_value
        except:
            return date_value

    def make_calculation_meta(self, stach_content_organization=None, format=None):
        '''
        stach_content_organization - [None, Row, Column, SimplifiedRow] default:SimplifiedRow
        format = [ JsonStach, Table, Tableau, BinaryStach, Bison, Binary, Pdf, Pptx, Feather ] default: JsonSTACH
        '''

        return CalculationMeta(
            stach_content_organization=stach_content_organization, format=format
        )

    def format_absolute_date(self, date_value):
        "Format date to YYYYMMDD format"
        try:
            return parse(date_value).strftime("%Y%m%d")
        except:
            return date_value

    def print_args(self, dict_args, as_sdk):
        '''
        Display Request Body as SDK JSON vs Web Request Equivalent
        '''

        additional_args = []
        data = []
        
        for key, val in dict_args.items():
            if "root" in key or key == "data":
                data.append(
                    self.__api_client.sanitize_for_serialization(val)
                    if not as_sdk
                    else val.to_dict()
                )
            else:
                additional_args.append(
                    f"{key.replace('_','-') if not as_sdk else key}: {val}"
                )
        for d in data:
            print(json.dumps(d, indent=5))

        if additional_args:
            print(
                "\nAdditional Arguments:" if as_sdk else "\nAdditional Request Headers:"
            )
            print("\n".join(additional_args) + "\n")
