import requests
import json
import time
from IPython.display import HTML, display
import urllib3
from fds.protobuf.stach.extensions.StachVersion import StachVersion
from fds.protobuf.stach.extensions.StachExtensionFactory import StachExtensionFactory
import pandas as pd
from enum import Enum

class Orientation(Enum):
    SIDE_BY_SIDE = 1
    STACKED = 2 
    COMMENTARY_ONLY = 3 



class PACommentary:

    __host = "https://api.factset.com/analytics/engines/pa/v3/calculations"

    

    __commentary_css_sbs = """
      <style type="text/css">
      .header {margin-top:-.75%;text-align:center;color:#00447c; font-weight:bold; font-size: 22px; margin-bottom:-.75%';
      }
     .date{
       text-align:center;color:#00447c;margin-top:-.75%;line-height:1; font-weight:bold;font-size: 20px;
     }     
      .body {display: block;font-size: 18px;color: #4b4b4b;background-color: #FFFFFF;font-family: arial, helvetica, sans-serif; padding:10px;}
      .section-header{display:block;font-size: 20px; color: #002856; font-weight:bold; margin-top:10px;} 
      .sub-section-header{display:block;font-size: 19px; color: #00447c; font-weight:bold; margin-top:10px;}


      .section{display:block; margin-left:15px; margin-top:5px;}
       </style>
    """
    __commentary_css = """
      <style type="text/css">
      .header {text-align:center;color:#00447c; font-weight:bold; font-size: 22px; margin-bottom:-.75%';
      }
     .date{
       text-align:center;color:#00447c;margin-top:-.75%;line-height:1; font-weight:bold;font-size: 20px;
     }     
      .body {display: block;font-size: 18px;color: #4b4b4b;background-color: #FFFFFF;font-family: arial, helvetica, sans-serif; padding:10px;}
      .section-header{display:block;font-size: 20px; color: #002856; font-weight:bold; margin-top:10px;} 
      .sub-section-header{display:block;font-size: 19px; color: #00447c; font-weight:bold; margin-top:10px;}


      .section{display:block; margin-left:15px; margin-top:5px;}
       </style>
    """

    def __init__(self, fds_username_hyphen_serial, fds_apikey, auth_token=None):
        urllib3.disable_warnings()
        self.__fds_username_hyphen_serial = fds_username_hyphen_serial
        self.__fds_apikey = fds_apikey
        self.__http_headers = {
            "Content-type": "application/json",
            "Accept": "application/json",
        }
        if auth_token:
            self.__http_headers["X-DataDirect-Auth-Token"] = auth_token

    def create_content(self, body: str):
        return f"<span class='section'>{body}</span>"

    def create_header(self, metadata: dict):
        title =  f"{metadata['tileName']}"
        VS = f"{metadata['portfolioDescription']} vs. {metadata['benchmarkDescription']}"
        dates = f"{metadata['startDate']['yyyymmdd']} to {metadata['endDate']['yyyymmdd']}"

        return f"<p class='header'> {title}</br>{VS}</br>{dates}</p>" 


    def create_html(self, commentary_data: dict, calculation_data:dict = {}, columns_to_rename: dict = {}, orientation:Orientation = Orientation.SIDE_BY_SIDE ):
        html = ""
        #Create Commentary HTML 
        has_header = False
        for section in commentary_data["sections"]:
            html += self.create_section_html(section,False, has_header)
            has_header=True        
        
        if orientation == Orientation.SIDE_BY_SIDE:
            pa_calc_html = f"<div class='column' style='flex: .5; padding: 10px;'>{self.format_pa_calc_results(self.return_calculation_results_as_dataframe(calculation_data)[0],columns_to_rename).to_html()}</div>"
            commentary_html = f"<div class='column' style='flex: .5; padding: 10px;'>{self.__commentary_css_sbs} <span class='body'>{html}</span></div>".replace("'", '"')

            style = 'flex: .5; padding: 10px;' 
            return f"<div class='row' style='display: flex; margin-top:-.5%'>{pa_calc_html}{commentary_html}</div>"    

        elif orientation == Orientation.STACKED:
            pa_calc_html = f"<div class='column' style='padding: 10px;'>{self.format_pa_calc_results(self.return_calculation_results_as_dataframe(calculation_data)[0],columns_to_rename).to_html()}</div>"
            commentary_html = f"<div class='column' style='padding: 10px;'>{self.__commentary_css} <span class='body'>{html}</span></div>".replace("'", '"')
            return f"<div class='row'; margin-top:-.5%'>{commentary_html}{pa_calc_html}</div>"
        else:
            style = 'padding: 10px;'
            return f"{self.__commentary_css}<div class='column' style={style}><span class='body'>{html}</span></div>".replace("'", '"')


    def create_section_html(self, section, is_subSection=False, has_header = True):
        html = ""
        header = ""
        for key, value in section.items():
            if key == "title":
                html += self.create_title(value, is_subSection)
            elif key == "content":
                html += self.create_content(value)
            elif key == "subSections":
                html += self.create_sub_sections(value)
            elif key == "metadata" and not has_header:
                header = self.create_header(value)
                has_header = True
                
        return header + html

    def create_sub_sections(self, subsections):
        inner_html = ""
        for subsection in subsections:
            inner_html += self.create_section_html(subsection, True)

        return f"<span class='section'>{inner_html}</span>"

  

    def create_title(self, title: str, is_subsesction: str = False):
        return f"<span class={'sub-section-header' if is_subsesction else 'section-header'}>{title}</span>"



    def display_as_html(self,pa_engine_response, pa_commentary_response: str, columns_to_rename: dict = {'total0':'Sector','group1':'Industry','group2':'Company'},orientation: Orientation = Orientation.SIDE_BY_SIDE):
        display(HTML(self.create_html(pa_commentary_response['data'],pa_engine_response['data'], columns_to_rename,orientation)))


    def display_commentary_as_html(self, pa_commentary_response: str):
        display(HTML(self.create_html(pa_commentary_response['data'],None,None,Orientation.COMMENTARY_ONLY)))



    def format_dataframe(self, df):
        '''Reindex and Center Headers'''
        header_align = "center"
        text_align = "center"
        return (df.style.hide(axis="index")).set_table_styles(
            [
                dict(selector="td", props=[("text-align", text_align)]),
                dict(selector="th", props=[("text-align", header_align)]),
                
            ]
        ).format(precision=3)

    def get_commentary_result(self, calculation_id: str, unit_id: str):
        """
        calculation_id: PA Engine API Calculation ID
        unit_id: PA Engine API Calulation Unit ID
        Expected Response: 200 (Short Running)
        """

        try:
            URL = f"{self.__host}/{calculation_id}/units/{unit_id}/commentary/result"
            print(
                f"Return PA Commentary Result for Calculation ID: {calculation_id} and Unit ID: {unit_id}"
            )
            print(f"GET: {URL}")

            response = self.__http_get(URL)

            print(f"Status Code: {response.status_code}")

            return response.json(), response.status_code, response.headers

        except Exception as e:
            print(f"Failed Returning Commentary Result - {response.status_code} - {e}")

    def get_commentary_status(self, calculation_id: str, unit_id: str, print_status = True):
        """
        calculation_id: PA Engine API Calculation ID
        unit_id: PA Engine API Calulation Unit ID
        Expected Response: 200 Completed
                            202 (Long Running)
        """

        try:
            URL = f"{self.__host}/{calculation_id}/units/{unit_id}/commentary/status"
            print(
                f"Check PA Commentary Status for Calculation ID: {calculation_id} and Unit ID: {unit_id}"
            )
            print(f"GET: {URL}")

            response = self.__http_get(URL)
            if print_status:
                print(f"Status Code: {response.status_code}")
            return response.json(), response.status_code, response.headers

        except Exception as e:
            print(f"Failed Checking Commentary Status - {response.status_code} - {e}")

    
    def get_pa_calculation_results(self, calculation_id: str, unit_id: str, print_status = True):
        try:
            URL = f"{self.__host}/{calculation_id}/units/{unit_id}/result"
            print(
                f"Return PA Calculation Response"
            )                     
            print(f"GET: {URL}")
            response = self.__http_get(URL)
            if print_status:
                print(f"Status Code: {response.status_code}")
            return response.json(), response.status_code, response.headers
        except Exception as e:
            print(f"Failed Returning PA Calculation Response - {response.status_code} - {e}")

     
    def return_calculation_results_as_dataframe(self, result):
        """Convert Row Based Stach into DataFrame (List)"""
        stachBuilder = StachExtensionFactory.get_row_organized_builder(
            StachVersion.V2
        )
        stachExtension = stachBuilder.set_package(result).build()
        dataFramesList = stachExtension.convert_to_dataframe()
        return dataFramesList
    
    
    def request_pa_commentary(self, calculation_id: str, unit_id: str, print_status = True ):
        """
        calculation_id: PA Engine API Calculation ID
        unit_id: PA Engine API Calulation Unit ID
        Expected Response: 201 (Short Running)
                            202 (Long Running)
        """
        """
        calculation_id: PA Engine API Calculation ID
        unit_id: PA Engine API Calulation Unit ID
        Expected Response: 201 (Short Running)
                            202 (Long Running)             
        """
        try:

            URL = f"{self.__host}/{calculation_id}/units/{unit_id}/commentary"
            print(
                f"Request PA Commentary for Calculation ID: {calculation_id} and Unit ID: {unit_id}"
            )
            print(f"POST: {URL}")

            response = self.__http_post(URL)
            if response.status_code<299:
                if print_status:
                    print(f"Status Code: {response.status_code} - Commentary Requested")
                    if response.status_code == 202:
                        print(f"Status URL {response.headers['location']}")
                return response.json(), response.status_code, response.headers
            
            else:
                if print_status:
                    print(f"Status Code: {response.status_code} - Failed to Request Commentary")
                    print(json.dumps(json.loads(response.text) ,indent=5))
                
                return json.loads(response.text),response.status_code,response.headers


        except Exception as e:
            print(f"Failed to Request Commentary - {response.status_code} - {e}")
            return response.text,response.status_code,response.headers

    def format_pa_calc_results(self, df, columns_to_rename: dict = None, remove_symbol_column=True):
        columns =  [l for l in df.columns.to_list() if " | " in l ]
        indexes =   [l for l in df.columns.to_list() if not " | " in l]
        sort_order = [True for l in df.columns.to_list() if not " | " in l]
        
        if remove_symbol_column:
            del indexes[0]
            del sort_order[0]
        unpivoted  = pd.melt(df,id_vars=indexes,value_vars=columns)
        list_columns, list_indexes, list_remove,list_sort_order = self.return_columns(unpivoted['variable'][0])
        unpivoted[list_columns] = unpivoted['variable'].str.split(pat = ' | ', expand=True, regex=False)            
        indexes = list_indexes + indexes
        sort_order = list_sort_order +sort_order
        unpivoted['Column'] = unpivoted['Column'].str.replace('+',' ')        
        unpivoted = unpivoted.drop(columns=list_remove)            
        unpivoted = unpivoted.pivot(index=indexes,columns=['Column'],values="value")
        unpivoted = unpivoted.reset_index()
        unpivoted.sort_values(indexes ,ascending=sort_order)
        unpivoted =  unpivoted.rename(columns=columns_to_rename) if columns_to_rename else unpivoted
        #return unpivoted
        
        return self.format_dataframe(unpivoted)


    def return_columns(self, string_columns: str, show_vs = False):
        list_headings = []
        list_remove = ["variable"]
        list_indexes = []
        list_sort_order = []
        list_columns = string_columns.split(" | ")
        for x , column in enumerate(list_columns):
            if " to " in column:
                list_headings.append("Date Range")
                list_indexes.append("Date Range")
                list_sort_order.append(False)
            elif " vs. " in column:
                list_headings.append("VS")
                if show_vs:
                    list_indexes.append("VS")
                else:
                    list_remove.append("VS")
            elif x < len(list_columns)-1:
                list_remove.append("Section")
                list_headings.append("Section")
            else:
                list_headings.append("Column")

        return list_headings,list_indexes, list_remove,list_sort_order


    def return_pa_calculation_results(self, calculation_id: str, unit_id: str, print_status = True):
        results,status,headers = self.get_pa_calculation_results(calculation_id,unit_id)
        return  self.format_pa_calc_results(self.return_calculation_results_as_dataframe(results['data'])[0])

    def return_pa_commentary(self, calculation_id: str, unit_id: str):
        """
        calculation_id: PA Engine API Calculation ID
        unit_id: PA Engine API Calulation Unit ID
        Expected Response: 200 Completed

        """
        response, status_code, headers = self.request_pa_commentary(
            calculation_id=calculation_id, unit_id=unit_id,print_status=False
        )
        if status_code == 202:
            response, status_code, headers = self.wait_for_long_running_request(
                calculation_id=calculation_id, unit_id=unit_id
            )
            return self.get_commentary_result(calculation_id, unit_id)

        if status_code == 201:
            return (response, status_code, headers)

        else:
            print(response)
            return response, status_code,headers


    def wait_for_long_running_request(self, calculation_id: str, unit_id: str):
        try:
            status_code = 202
            max_age = 5
            while status_code == 202:
                print(f"Status Code: {status_code} - Sleeping for {max_age} seconds")
                time.sleep(int(max_age))
                response, status_code, headers = self.get_commentary_status(
                    calculation_id, unit_id,False
                )
                

            if status_code == 200:
                print(f"Status Code: {status_code}")
                print("Final Response Body:")
                print(json.dumps(response, indent=5))
                return response, status_code, headers

        except Exception as e:
            print(e)

    # get web request
    def __http_get(self, URL):
        return requests.get(
            URL,
            auth=(self.__fds_username_hyphen_serial, self.__fds_apikey),
            headers=self.__http_headers,
            verify=False,
        )

    # post web request
    def __http_post(self, URL):
        #try:

            return requests.post(
                URL,
                auth=(self.__fds_username_hyphen_serial, self.__fds_apikey),
                headers=self.__http_headers,
                verify=False,
            )
        #except Exception as e:
        #    return str(e)


